@@ -43,12 +43,16 @@
 
 #define GDM_GREETER_SESSION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GDM_TYPE_GREETER_SESSION, GdmGreeterSessionPrivate))
 
+#define MAX_LOGIN_TRIES 3
+
 struct GdmGreeterSessionPrivate
 {
         GdmGreeterClient      *client;
 
         GtkWidget             *login_window;
         GtkWidget             *panel;
+
+        guint                  num_tries;
 };
 
 enum {
@@ -98,18 +102,41 @@ on_reset (GdmGreeterClient  *client,
 {
         g_debug ("GdmGreeterSession: Reset");
 
-        gdm_greeter_panel_reset (GDM_GREETER_PANEL (session->priv->panel));
+        session->priv->num_tries = 0;
+
         gdm_greeter_login_window_reset (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window));
 }
 
+static void
+on_authentication_failed (GdmGreeterClient  *client,
+                          GdmGreeterSession *session)
+{
+        g_debug ("GdmGreeterSession: Authentication failed");
+
+        session->priv->num_tries++;
+
+        if (session->priv->num_tries < MAX_LOGIN_TRIES) {
+                g_debug ("GdmGreeterSession: Retrying login (%d)",
+                         session->priv->num_tries);
+
+                gdm_greeter_login_window_authentication_failed (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window));
+        } else {
+                g_debug ("GdmGreeterSession: Maximum number of login tries exceeded (%d) - resetting",
+                         session->priv->num_tries - 1);
+                session->priv->num_tries = 0;
+
+                gdm_greeter_login_window_reset (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window));
+        }
+}
+
 static void
 show_or_hide_user_options (GdmGreeterSession *session,
                            const char        *username)
 {
     if (username != NULL && strcmp (username, GDM_USER_CHOOSER_USER_OTHER) != 0) {
-            gdm_greeter_panel_show_user_options (GDM_GREETER_PANEL (session->priv->panel));
+            //gdm_greeter_panel_show_user_options (GDM_GREETER_PANEL (session->priv->panel));
     } else {
-            gdm_greeter_panel_hide_user_options (GDM_GREETER_PANEL (session->priv->panel));
+            //gdm_greeter_panel_hide_user_options (GDM_GREETER_PANEL (session->priv->panel));
     }
 }
 
@@ -128,18 +155,6 @@ on_default_language_name_changed (GdmGreeterClient  *client,
                                   GdmGreeterSession *session)
 {
         g_debug ("GdmGreeterSession: default language name changed: %s", text);
-        gdm_greeter_panel_set_default_language_name (GDM_GREETER_PANEL (session->priv->panel),
-                                                     text);
-}
-
-static void
-on_default_layout_name_changed (GdmGreeterClient  *client,
-                                const char        *text,
-                                GdmGreeterSession *session)
-{
-        g_debug ("GdmGreeterSession: default layout name changed: %s", text);
-        gdm_greeter_panel_set_default_layout_name (GDM_GREETER_PANEL (session->priv->panel),
-                                                   text);
 }
 
 static void
@@ -148,8 +163,6 @@ on_default_session_name_changed (GdmGreeterClient  *client,
                                  GdmGreeterSession *session)
 {
         g_debug ("GdmGreeterSession: default session name changed: %s", text);
-        gdm_greeter_panel_set_default_session_name (GDM_GREETER_PANEL (session->priv->panel),
-                                                    text);
 }
 
 static void
@@ -240,20 +253,6 @@ on_select_language (GdmGreeterSession     *session,
                                                  text);
 }
 
-static void
-on_select_layout (GdmGreeterSession      *session,
-                  const char             *text)
-{
-        gdm_greeter_client_call_select_layout (session->priv->client,
-                                               text);
-}
-
-static void
-on_dialog_hidden (GdmGreeterSession     *session)
-{
-        gtk_window_present (GTK_WINDOW (session->priv->login_window));
-}
-
 static void
 on_select_user (GdmGreeterLoginWindow *login_window,
                 const char            *text,
@@ -268,7 +267,6 @@ static void
 on_cancelled (GdmGreeterLoginWindow *login_window,
               GdmGreeterSession     *session)
 {
-        gdm_greeter_panel_hide_user_options (GDM_GREETER_PANEL (session->priv->panel));
         gdm_greeter_client_call_cancel (session->priv->client);
 }
 
@@ -291,8 +289,8 @@ get_tallest_monitor_at_point (GdkScreen *screen,
                               int        x,
                               int        y)
 {
-        GdkRectangle area;
-        GdkRegion *region;
+        cairo_rectangle_int_t area;
+        cairo_region_t *region;
         int i;
         int monitor;
         int n_monitors;
@@ -303,15 +301,15 @@ get_tallest_monitor_at_point (GdkScreen *screen,
         monitor = -1;
         for (i = 0; i < n_monitors; i++) {
                 gdk_screen_get_monitor_geometry (screen, i, &area);
-                region = gdk_region_rectangle (&area);
+                region = cairo_region_create_rectangle (&area);
 
-                if (gdk_region_point_in (region, x, y)) {
+                if (cairo_region_contains_point (region, x, y)) {
                         if (area.height > tallest_height) {
                                 monitor = i;
                                 tallest_height = area.height;
                         }
                 }
-                gdk_region_destroy (region);
+                cairo_region_destroy (region);
         }
 
         if (monitor == -1) {
@@ -347,21 +345,11 @@ toggle_panel (GdmGreeterSession *session,
                                           G_CALLBACK (on_select_language),
                                           session);
 
-                g_signal_connect_swapped (session->priv->panel,
-                                          "layout-selected",
-                                          G_CALLBACK (on_select_layout),
-                                          session);
-
                 g_signal_connect_swapped (session->priv->panel,
                                           "session-selected",
                                           G_CALLBACK (on_select_session),
                                           session);
 
-                g_signal_connect_swapped (session->priv->panel,
-                                          "dialog-hidden",
-                                          G_CALLBACK (on_dialog_hidden),
-                                          session);
-
                 gtk_widget_show (session->priv->panel);
         } else {
                 gtk_widget_destroy (session->priv->panel);
@@ -579,6 +567,10 @@ gdm_greeter_session_init (GdmGreeterSession *session)
                           "reset",
                           G_CALLBACK (on_reset),
                           session);
+        g_signal_connect (session->priv->client,
+                          "authentication-failed",
+                          G_CALLBACK (on_authentication_failed),
+                          session);
         g_signal_connect (session->priv->client,
                           "selected-user-changed",
                           G_CALLBACK (on_selected_user_changed),
@@ -587,10 +579,6 @@ gdm_greeter_session_init (GdmGreeterSession *session)
                           "default-language-name-changed",
                           G_CALLBACK (on_default_language_name_changed),
                           session);
-        g_signal_connect (session->priv->client,
-                          "default-layout-name-changed",
-                          G_CALLBACK (on_default_layout_name_changed),
-                          session);
         g_signal_connect (session->priv->client,
                           "default-session-name-changed",
                           G_CALLBACK (on_default_session_name_changed),
